#!/bin/sh
#
# Copyright (c) 2021  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

set -e

remrun_version='0.1.0'

version()
{
	echo "remrun $remrun_version"
}

usage()
{
	cat <<'EOUSAGE'
Usage:	remrun [-Nsv] [--noop] [--stdin] [--verbose]
		[-C local_cksum_command] [-c remote_cksum_command]
		hostspec /path/to/file
	remrun -h | --help | -V | --version | --features

	-C	specify the local command to run instead of sha256sum
	-c	specify the remote command to run instead of sha256sum
	-h	display program usage information and exit
	-N	no-operation mode; do not execute the file on the remote host
	-s	let the remote command read from the standard input stream
	-V	display program version information and exit
	-v	verbose mode; display diagnostic output
EOUSAGE
}

features()
{
	echo "Features: remrun=$remrun_version hostspec=1.0"
}

diag()
{
	[ -z "$v" ] || printf -- '%s\n' "$*" 1>&2
}

create_remote_tempfile()
{
	local hostspec="$1"
	local tempf

	diag "About to create a temporary file at '$hostspec'"
	if ! tempf="$(ssh -n -- "$hostspec" mktemp remrun.XXXXXX)"; then
		echo "Could not run mktemp at '$hostspec' via SSH" 1>&2
		exit 1
	fi
	diag "Got '$tempf' as the remote temporary filename"
	printf -- '%s\n' "$tempf"
}

remove_remote_tempfile()
{
	local hostspec="$1" remtempf="$2"

	diag "About to remove the remote temporary file '$remtempf'"
	if ! ssh -n -- "$hostspec" rm -- "$remtempf"; then
		echo "Could not remove the '$remtempf' temporary file at '$hostspec'" 1>&2
		exit 1
	fi
}

store_and_run()
{
	local hostspec="$1" remtempf="$2" filename="$3" cksum_local="$4"
	local cksum_remote remtempfull nopt

	diag "Storing the contents of '$filename' into '$hostspec:$remtempf'"
	if ! ssh -- "$hostspec" sh -c "cat > '$remtempf'" < "$filename"; then
		echo "Could not write '$filename' to '$remtempf' at '$hostspec'" 1>&2
		exit 1
	fi

	diag "Obtaining the checksum"
	if ! cksum_remote="$(ssh -n -- "$hostspec" "$remote_cksum_cmd -- $remtempf")"; then
		echo "Could not obtain the checksum of '$remtempf' at '$hostspec'" 1>&2
		exit 1
	fi
	cksum_remote="${cksum_remote%% *}"
	diag "Obtained remote checksum $cksum_remote"
	if [ "$cksum_local" != "$cksum_remote" ]; then
		echo "The remote checksum '$cksum_remote' does not match the local one '$cksum_local'" 1>&2
		return 1
	fi

	diag "Setting the access permissions on the remote file"
	if ! ssh -n -- "$hostspec" chmod -- 700 "$remtempf"; then
		echo "Failed to set the access permissions of '$remtempf' at '$hostspec'" 1>&2
		exit 1
	fi

	remtempfull="$remtempf"
	if [ "${remtempfull%%/*}" = "$remtempfull" ]; then
		diag "Prepending ./ to the name of the remote file"
		remtempfull="./$remtempfull"
	fi
	diag "Executing the remote file $remtempfull"
	if [ -n "$stdin" ]; then
		diag "Not passing '-n' to the SSH client"
		nopt=''
	else
		nopt='-n'
	fi
	# shellcheck disable=2086,2248  # both $noop and $nopt are under our control
	if ! ssh $nopt -- "$hostspec" $noop "$remtempfull"; then
		echo "Running '$remtempfull' at '$hostspec' failed" 1>&2
		exit 1
	fi
}

doit()
{
	local hostspec="$1" filename="$2"
	local cksum_local remtempf tempf_stdin='/nonexistent'

	if [ "$filename" = '-' ]; then
		tempf_stdin="$(mktemp remrun-stdin.XXXXXX)"
		# shellcheck disable=SC2064
		trap "rm -f -- '$tempf_stdin'" HUP INT EXIT QUIT TERM

		diag "Reading the standard input stream into $tempf_stdin"
		cat > "$tempf_stdin"

		diag "Processing $tempf_stdin now"
		filename="$tempf_stdin"
	fi

	if [ ! -f "$filename" ] || [ ! -r "$filename" ]; then
		echo "Not a readable regular file: $filename" 1>&2
		exit 1
	fi
	diag "Obtaining the local checksum of $filename"
	cksum_local="$($local_cksum_cmd -- "$filename")"
	cksum_local="${cksum_local%% *}"
	if [ -z "$cksum_local" ]; then
		echo "The checksum of $filename cannot be an empty string" 1>&2
		exit 1
	fi
	diag "Obtained local checksum $cksum_local"

	remtempf="$(create_remote_tempfile "$hostspec")"
	# shellcheck disable=SC2064
	trap "remove_remote_tempfile '$hostspec' '$remtempf'; rm -f -- '$tempf_stdin'" HUP INT EXIT QUIT TERM
	store_and_run "$hostspec" "$remtempf" "$filename" "$cksum_local"
}

unset hflag noop show_features stdin Vflag v
local_cksum_cmd='sha256sum'
remote_cksum_cmd='sha256sum'

while getopts 'C:c:hNsVv-:' o; do
	case "$o" in
		C)
			local_cksum_cmd="$OPTARG"
			;;

		c)
			remote_cksum_cmd="$OPTARG"
			;;

		h)
			hflag=1
			;;

		N)
			noop='echo'
			;;

		s)
			stdin=1
			;;

		V)
			Vflag=1
			;;

		v)
			v='-v'
			;;

		-)
			case "$OPTARG" in
				features)
					show_features=1
					;;

				help)
					hflag=1
					;;

				noop)
					noop='echo'
					;;

				stdin)
					stdin=1
					;;

				verbose)
					v='-v'
					;;

				version)
					Vflag=1
					;;

				*)
					echo "Invalid long option '$OPTARG' specified" 1>&2
					usage 1>&2
					exit 1
					;;
			esac
			;;

		*)
			usage 1>&2
			exit 1
			;;
	esac
done
[ -z "$Vflag" ] || version
[ -z "$hflag" ] || usage
[ -z "$show_features" ] || features
[ -z "$Vflag$hflag$show_features" ] || exit 0

shift "$((OPTIND - 1))"
if [ "$#" -ne 2 ]; then
	usage 1>&2
	exit 1
fi

if [ -n "$stdin" ] && [ "$2" = '-' ]; then
	echo 'The -s option may not be used with "-" as a filename' 1>&2
	exit 1
fi

doit "$1" "$2"
